/*
 * Copyright 2024 Arne Vajh�j.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.vms.tig.client;

import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;

import dk.vajhoej.vms.tig.Arg;
import dk.vajhoej.vms.tig.Arg.DataType;
import dk.vajhoej.vms.tig.Arg.Usage;
import dk.vajhoej.vms.tig.Method;
import dk.vajhoej.vms.tig.shared.Util;

public class CSharpClientGen implements ClientGen {
	private PrintWriter pw;
	@Override
	public String fileName(String imgnam) {
		return className(imgnam) + ".cs";
	}
	@Override
	public String className(String imgnam) {
		return Util.capitalize(imgnam) + "Client";
	}
	@Override
	public String typeName(Arg.DataType typ) {
		switch(typ) {
			case Byte:
				return "byte";
			case Word:
				return "short";
			case LongWord:
				return "int";
			case QuadWord:
				return "long";
			case F_Float:
			case S_Float:
				return "float";
			case D_Float:
			case G_Float:
			case T_Float:
				return "double";
			case FixedCharacterString:
			case VariableCharacterString:
			case NullTermCharacterString:
				return "string";
			case Block:
				return "Block";
			default:
				throw new RuntimeException("Unsupported type: " + typ);
		}
	}
	@Override
	public void open(String fnm) throws IOException {
		pw = new PrintWriter(new FileWriter(fnm));
	}
	private boolean usens = false;
	@Override
	public void genClassHeader(String ns, String clznam, String ooctx, String thisbase) {
		pw.printf("// GENERATED CODE - edits may be overwritten\n");
		pw.printf("using System;\n");
		pw.printf("using System.IO;\n");
		pw.printf("using System.Net.Sockets;\n");
		pw.printf("\n");
		if(thisbase != null) {
			pw.printf("using Vajhoej.Record;\n");
		}
		pw.printf("using Vajhoej.VMS.TIG;\n");
		pw.printf("\n");
		if(ns != null) {
			pw.printf("namespace %s\n", ns);
			pw.printf("{\n");
			pw.printf("\n");
			usens = true;
		}
		if(thisbase != null) {
			pw.printf("[Struct]\n");
		}
		pw.printf("public class %s\n", clznam);
		if(thisbase != null) {
			pw.printf("    : %s\n", Util.capitalize(thisbase));
		}
		pw.printf("{\n");
		if(thisbase != null) {
			pw.printf("    [TransientField]\n");
		}
		pw.printf("    private string host;\n");
		if(thisbase != null) {
			pw.printf("    [TransientField]\n");
		}
		pw.printf("    private int port;\n");
		if(ooctx != null) {
			if(!Character.isDigit(ooctx.charAt(0))) {
				pw.printf("    private %s _ooctx = new %s();\n", Util.capitalize(ooctx), Util.capitalize(ooctx));
			} else {
				pw.printf("    private byte[] _ooctx = new byte[%d];\n", Integer.parseInt(ooctx));
			}
		}
	}
	@Override
	public void genCtor(String clznam, int port, String ooctx, String thisbase) {
		pw.printf("    public %s(string host)\n", clznam);
		pw.printf("    {\n");
		pw.printf("        this.host = host;\n");
		pw.printf("        this.port = %d;\n", port);
		pw.printf("    }\n");
		pw.printf("    public %s()\n", clznam);
		pw.printf("    {\n");
		pw.printf("        // just needed for temp instance\n");
		pw.printf("    }\n");
		if(ooctx != null && !Character.isDigit(ooctx.charAt(0))) {
			pw.printf("    public %s OOContext\n",  Util.capitalize(ooctx));
			pw.printf("    {\n");
			pw.printf("        get { return _ooctx; }\n");
			pw.printf("    }\n");
		}
	}
	@Override
	public void genMethodHeader(String methnam, Method.ArgList args) {
		pw.printf("    public int %s(", Util.capitalize(methnam));
		boolean first = true;
		for(Arg arg : args.getArgs()) {
			if(first) {
				first = false;
			} else {
				pw.printf(", ");
			}
			if(arg.getUse() != Usage.In) {
				// make In and InOut argument ref
				pw.printf("ref ");
			}
			pw.printf("%s", arg.getType() != Arg.DataType.Block ? typeName(arg.getType()) : arg.getImpl());
			pw.printf(" %s", arg.getName());
		}
		pw.printf(")\n");
		pw.printf("    {\n");
		pw.printf("        TcpClient tc = new TcpClient(host, port);\n");
		pw.printf("        Connection con = new Connection(tc);\n");
		pw.printf("        con.Prep();\n");
		pw.printf("        con.PutString(\"%s\");\n", methnam);
	}
	@Override
	public void genMethodBody(String ooctx, String thisbase, Method.ArgList args) {
		if(ooctx != null) {
			if(!Character.isDigit(ooctx.charAt(0))) {
				pw.printf("        con.PutBlock(_ooctx);\n");
			} else {
				pw.printf("        con.PutBytes(_ooctx);\n");
			}
		}
		if(thisbase != null) {
			pw.printf("        con.PutBlock(this);\n");
		}
		for(Arg arg : args.getArgs()) {
			if(arg.getUse() != Usage.Out || arg.getType() == DataType.Block) {
		    	// for In and InOut write value to connection
		        pw.printf("        con.Put%s(%s);\n", Util.capitalize(typeName(arg.getType())), arg.getName());
			}
		}
		pw.printf("        con.Send();\n");
		pw.printf("        con.Recv();\n");
		pw.printf("        int retval = con.GetInt();\n");
		if(ooctx != null) {
			if(!Character.isDigit(ooctx.charAt(0))) {
				pw.printf("        _ooctx = (%s)con.GetBlock<%s>(typeof(%s));\n", Util.capitalize(ooctx), Util.capitalize(ooctx), Util.capitalize(ooctx));
			} else {
				pw.printf("        _ooctx = con.GetBytes();\n");
			}
		}
		if(thisbase != null) {
			pw.printf("        con.GetBlock(this);\n");
		}
		for(Arg arg : args.getArgs()) {
			if(arg.getUse() != Usage.In) {
				if(arg.getType() != DataType.Block) {
					// for InOut and Out non-block read value from connection
					pw.printf("        %s = con.Get%s();\n", arg.getName(), Util.capitalize(typeName(arg.getType())));
				} else {
					// for InOut and Out non-block read value from connection (class is needed to properly deserialize)
					pw.printf("        %s = con.Get%s<%s>(typeof(%s));\n", arg.getName(), Util.capitalize(typeName(arg.getType())), arg.getImpl(), arg.getImpl());
				}
			}
		}
	}
	@Override
	public void genMethodFooter(Method.ArgList args) {
		pw.printf("        con.Close();\n");
		pw.printf("        return retval;\n");
		pw.printf("    }\n");
	}
	@Override
	public void genClassFooter() {
		pw.printf("}\n");
		if(usens) {
			pw.printf("\n");
			pw.printf("}\n");
		}
	}
	@Override
	public void close() {
		pw.close();
	}
}
