/*
 * Copyright 2024 Arne Vajh�j.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.vms.tig.client;

import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;

import dk.vajhoej.vms.tig.Arg;
import dk.vajhoej.vms.tig.Arg.DataType;
import dk.vajhoej.vms.tig.Arg.Usage;
import dk.vajhoej.vms.tig.shared.Util;
import dk.vajhoej.vms.tig.Method;

public class CppClientGen implements ClientGen {
	private PrintWriter pw_cpp;
	private PrintWriter pw_h;
	private PrintWriter pw_exh;
	@Override
	public String fileName(String imgnam) {
		return className(imgnam) + ".cpp";
	}
	@Override
	public String className(String imgnam) {
		return Util.capitalize(imgnam) + "Client";
	}
	@Override
	public String typeName(Arg.DataType typ) {
		switch(typ) {
			case Byte:
				return "int8_t";
			case Word:
				return "int16_t";
			case LongWord:
				return "int32_t";
			case QuadWord:
				return "int64_t";
			case F_Float:
			case S_Float:
				return "float";
			case D_Float:
			case G_Float:
			case T_Float:
				return "double";
			case FixedCharacterString:
			case VariableCharacterString:
			case NullTermCharacterString:
				return "charptr";
			case Block:
				return "block";
			default:
				throw new RuntimeException("Unsupported type: " + typ);
		}
	}
	@Override
	public void open(String fnm) throws IOException {
		pw_cpp = new PrintWriter(new FileWriter(fnm));
		String fnm2 = fnm.substring(0, fnm.length() - 4) + ".h";
		pw_h = new PrintWriter(new FileWriter(fnm2));
		String fnm3 = fnm.substring(0, fnm.length() - 4) + "Ex.h";
		pw_exh = new PrintWriter(new FileWriter(fnm3));
	}
	private String prefix;
	@Override
	public void genClassHeader(String ns, String clznam, String ooctx, String thisbase) {
		if(thisbase != null) {
			System.out.println(">>>> thisbase is potentially dangerous for C++ (you need to account correctly for vtable) <<<<");
		}
		prefix = clznam;
		pw_exh.printf("// GENERATED CODE - edits may be overwritten\n");
		if(ooctx != null && !Character.isDigit(ooctx.charAt(0))) {
			pw_exh.printf("#include \"%s.h\"\n", ooctx.toLowerCase());
			pw_exh.printf("\n");
		}
		if(thisbase != null) {
			pw_exh.printf("#include \"%s.h\"\n", thisbase);
			pw_exh.printf("\n");
		}
		pw_h.printf("// GENERATED CODE - edits may be overwritten\n");
		pw_h.printf("#ifndef %s_H\n", prefix.toUpperCase());
		pw_h.printf("#define %s_H\n", prefix.toUpperCase());
		pw_h.printf("\n");
		pw_h.printf("typedef const char *charptr;\n");
		pw_h.printf("\n");
		pw_h.printf("#include \"%sEx.h\"\n", prefix);
		pw_h.printf("\n");
		pw_h.printf("class %s\n", clznam);
		if(thisbase != null) {
			pw_h.printf("    : public %s\n", Util.capitalize(thisbase));
		}
		pw_h.printf("{\n");
		pw_h.printf("private:\n");
		pw_h.printf("    const char* host;\n");
		pw_h.printf("    int port;\n");
		pw_h.printf("    struct con_t *con;\n");
		if(ooctx != null) {
			if(!Character.isDigit(ooctx.charAt(0))) {
				pw_h.printf("    %s _ooctx;\n", ooctx.toLowerCase());
			} else {
				pw_h.printf("    char _ooctx[%d];\n", Integer.parseInt(ooctx));
			}
		}
		pw_h.printf("public:\n");
		if(ooctx != null && !Character.isDigit(ooctx.charAt(0))) {
			pw_h.printf("    %s GetOOContext();\n",  ooctx.toLowerCase());
		}
		pw_cpp.printf("// GENERATED CODE - edits may be overwritten\n");
		pw_cpp.printf("#include <cstdint>\n");
		pw_cpp.printf("#include <cstring>\n");
		pw_cpp.printf("\n");
		pw_cpp.printf("#include \"tig.h\"\n");
		pw_cpp.printf("\n");
		pw_cpp.printf("#include \"%s.h\"\n", clznam);
		pw_cpp.printf("\n");
	}
	@Override
	public void genCtor(String clznam, int port, String ooctx, String thisbase) {
		pw_h.printf("    %s(const char* host);\n", clznam);
		pw_cpp.printf("%s::%s(const char* host)\n", clznam, clznam);
		if(thisbase != null) {
			pw_cpp.printf("    : %s()\n", thisbase);
		}
		pw_cpp.printf("{\n");
		pw_cpp.printf("    this->host = host;\n");
		pw_cpp.printf("    this->port = %d;\n", port);
		if(ooctx != null) {
			if(!Character.isDigit(ooctx.charAt(0))) {
				pw_cpp.printf("    _ooctx = new %s_t;\n", ooctx);
				pw_cpp.printf("    memset(_ooctx, 0, sizeof(*_ooctx));\n");
			} else {
				pw_cpp.printf("    memset(_ooctx, 0, sizeof(_ooctx));\n");
			}
		}
		pw_cpp.printf("}\n");
		if(ooctx != null && !Character.isDigit(ooctx.charAt(0))) {
			pw_cpp.printf("%s %s::GetOOContext()\n", ooctx.toLowerCase(), clznam);
			pw_cpp.printf("{\n");
			pw_cpp.printf("    return _ooctx;\n");
			pw_cpp.printf("}\n");
		}
		pw_cpp.printf("\n");
	}
	@Override
	public void genMethodHeader(String methnam, Method.ArgList args) {
		for(Arg arg : args.getArgs()) {
			if(arg.getType() == Arg.DataType.Block) {
				pw_exh.printf("#include \"%s.h\"\n", arg.getImpl().toLowerCase());
				pw_exh.printf("\n");
			}
		}
		pw_h.printf("    int32_t %s(", Util.capitalize(methnam));
		boolean hfirst = true;
		for(Arg arg : args.getArgs()) {
			if(hfirst) {
				hfirst = false;
			} else {
				pw_h.printf(", ");
			}
			pw_h.printf("%s", arg.getType() != Arg.DataType.Block ? typeName(arg.getType()) : arg.getImpl().toLowerCase());
			if(arg.getUse() != Usage.In) {
				// make In and InOut argument by pointer
				pw_h.printf("*");
			}
			pw_h.printf(" %s", arg.getName());
		}
		pw_h.printf(");\n");
		boolean cppfirst = true;
		pw_cpp.printf("int32_t %s::%s(", prefix, Util.capitalize(methnam));
		for(Arg arg : args.getArgs()) {
			if(cppfirst) {
				cppfirst = false;
			} else {
				pw_cpp.printf(", ");
			}
			pw_cpp.printf("%s", arg.getType() != Arg.DataType.Block ? typeName(arg.getType()) : arg.getImpl().toLowerCase());
			if(arg.getUse() != Usage.In) {
				// make In and InOut argument by pointer
				pw_cpp.printf("*");
			}
			pw_cpp.printf(" %s", arg.getName());
		}
		pw_cpp.printf(")\n");
		pw_cpp.printf("{\n");
		pw_cpp.printf("    con = con_init(this->host, this->port);\n");
		pw_cpp.printf("    con_prep(con);\n");
		pw_cpp.printf("    con_put_charptr(con, \"%s\");\n", methnam);
	}
	@Override
	public void genMethodBody(String ooctx, String thisbase, Method.ArgList args) {
		if(ooctx != null) {
			if(!Character.isDigit(ooctx.charAt(0))) {
				pw_cpp.printf("    con_put_block(con, _ooctx, sizeof(*_ooctx));\n");
			} else {
				pw_cpp.printf("    con_put_bytes(con, _ooctx, sizeof(_ooctx));\n");
			}
		}
		if(thisbase != null) {
			pw_cpp.printf("    con_put_bytes(con, (const char *)this, sizeof(%s));\n", thisbase);
		}
		for(Arg arg : args.getArgs()) {
			if(arg.getUse() != Usage.Out || arg.getType() == DataType.Block) {
				if(arg.getType() != DataType.Block) {
			    	// for In and InOut non-blocks write value to connection
					if(arg.getType() == DataType.Word) {
						pw_cpp.printf("    %s = to_little_16(%s);\n", arg.getName(), arg.getName());
					}
					if(arg.getType() == DataType.LongWord) {
						pw_cpp.printf("    %s = to_little_32(%s);\n", arg.getName(), arg.getName());
					}
					if(arg.getType() == DataType.QuadWord) {
						pw_cpp.printf("    %s = to_little_64(%s);\n", arg.getName(), arg.getName());
					}
			        pw_cpp.printf("    con_put_%s(con, %s);\n", typeName(arg.getType()).toLowerCase(), arg.getName());
				} else {
			    	// for In and InOut blocks write value to connection
			        pw_cpp.printf("    con_put_%s(con, *%s, sizeof(**%s));\n", typeName(arg.getType()).toLowerCase(), arg.getName(), arg.getName());
				}
			}
		}
		pw_cpp.printf("    con_send(con);\n");
		pw_cpp.printf("    con_recv(con);\n");
		pw_cpp.printf("    int retval = con_get_int32_t(con);\n");
		if(ooctx != null) {
			if(!Character.isDigit(ooctx.charAt(0))) {
				pw_cpp.printf("    _ooctx = (%s)con_get_block(con);\n", ooctx.toLowerCase());
			} else {
				pw_cpp.printf("    con_get_bytes(con, _ooctx, sizeof(_ooctx));\n");
			}
		}
		if(thisbase != null) {
			pw_cpp.printf("    con_get_bytes(con, (char *)this, sizeof(%s));\n", thisbase);
		}
		for(Arg arg : args.getArgs()) {
			if(arg.getUse() != Usage.In) {
				if(arg.getType() != DataType.Block) {
					// for InOut and Out non-block read value from connection
					pw_cpp.printf("    *%s = con_get_%s(con);\n", arg.getName(), typeName(arg.getType()).toLowerCase());
					if(arg.getType() == DataType.Word) {
						pw_cpp.printf("    *%s = from_little_16(*%s);\n", arg.getName(), arg.getName());
					}
					if(arg.getType() == DataType.LongWord) {
						pw_cpp.printf("    *%s = from_little_32(*%s);\n", arg.getName(), arg.getName());
					}
					if(arg.getType() == DataType.QuadWord) {
						pw_cpp.printf("    *%s = from_little_64(*%s);\n", arg.getName(), arg.getName());
					}
				} else {
					// for InOut and Out non-block read value from connection
					pw_cpp.printf("    *%s = (%s)con_get_%s(con);\n", arg.getName(), arg.getImpl().toLowerCase(), typeName(arg.getType()).toLowerCase());
				}
			}
		}
	}
	@Override
	public void genMethodFooter(Method.ArgList args) {
		pw_cpp.printf("    con_close(con);\n");
		pw_cpp.printf("    return retval;\n");
		pw_cpp.printf("}\n");
		pw_cpp.printf("\n");
	}
	@Override
	public void genClassFooter() {
		pw_h.printf("};\n");
	}
	@Override
	public void close() {
		pw_cpp.close();
		pw_h.printf("\n");
		pw_h.printf("#endif\n");
		pw_h.close();
		pw_exh.close();
	}
}
