/*
 * Copyright 2024 Arne Vajh�j.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.vms.tig.client;

import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;

import dk.vajhoej.vms.tig.Arg;
import dk.vajhoej.vms.tig.Arg.DataType;
import dk.vajhoej.vms.tig.Arg.Usage;
import dk.vajhoej.vms.tig.Method;
import dk.vajhoej.vms.tig.shared.Util;

public class PhpClientGen implements ClientGen {
	private PrintWriter pw;
	@Override
	public String fileName(String imgnam) {
		return className(imgnam) + ".php";
	}
	@Override
	public String className(String imgnam) {
		return Util.capitalize(imgnam) + "Client";
	}
	@Override
	public String typeName(Arg.DataType typ) {
		switch(typ) {
			case Byte:
				return "int8";
			case Word:
				return "int16";
			case LongWord:
				return "int32";
			case QuadWord:
				return "int64";
			case F_Float:
			case S_Float:
				return "float";
			case D_Float:
			case G_Float:
			case T_Float:
				return "double";
			case FixedCharacterString:
			case VariableCharacterString:
			case NullTermCharacterString:
				return "string";
			case Block:
				return "block";
			default:
				throw new RuntimeException("Unsupported type: " + typ);
		}
	}
	@Override
	public void open(String fnm) throws IOException {
		pw = new PrintWriter(new FileWriter(fnm));
	}
	@Override
	public void genClassHeader(String ns, String clznam, String ooctx, String thisbase) {
		pw.printf("<?php\n");
		pw.printf("\n");
		pw.printf("// GENERATED CODE - edits may be overwritten\n");
		pw.printf("\n");
		pw.printf("include 'tig.php';\n");
		if(ooctx != null && !Character.isDigit(ooctx.charAt(0))) {
			pw.printf("include '%s.php';\n", Util.capitalize(ooctx));
		}
		if(thisbase != null) {
			pw.printf("include '%s.php';\n", Util.capitalize(thisbase));
		}
		pw.printf("\n");
		if(thisbase != null) {
			pw.printf("class %s extends %s {\n", clznam, thisbase);
		} else {
			pw.printf("class %s {\n", clznam);
		}
	}
	@Override
	public void genCtor(String clznam, int port, String ooctx, String thisbase) {
		pw.printf("    private $host;\n");
		pw.printf("    private $port;\n");
		pw.printf("    private $con;\n");
		if(ooctx != null) {
			pw.printf("    private $_ooctx;\n");
		}
		pw.printf("    public function __construct($host) {\n", clznam);
		pw.printf("        $this->host = $host;\n");
		pw.printf("        $this->port = %d;\n", port);
		if(ooctx != null) {
			if(!Character.isDigit(ooctx.charAt(0))) {
				pw.printf("        $this->_ooctx = new %s();\n", Util.capitalize(ooctx));
				pw.printf("    }\n");
				pw.printf("    public function oocontext() {\n");
				pw.printf("        return $this->_ooctx;\n");
			} else {
				pw.printf("        $this->_ooctx = str_repeat('\0', %d)\n", Integer.parseInt(ooctx));
			}
		}
		pw.printf("    }\n");
	}
	@Override
	public void genMethodHeader(String methnam, Method.ArgList args) {
		pw.printf("    public function %s(", methnam);
		boolean first = true;
		for(Arg arg : args.getArgs()) {
			if(first) {
				first = false;
			} else {
				pw.printf(", ");
			}
			if(arg.getUse() == Usage.In) {
				pw.printf("$%s", arg.getName());
			} else {
				pw.printf("&$%s", arg.getName());
			}
		}
		pw.printf(") {\n");
		pw.printf("        $this->con = new Connection($this->host, $this->port);\n");
		pw.printf("        $this->con->prep();\n");
		pw.printf("        $this->con->put_string('%s');\n", methnam);
	}
	@Override
	public void genMethodBody(String ooctx, String thisbase, Method.ArgList args) {
		if(ooctx != null) {
			if(!Character.isDigit(ooctx.charAt(0))) {
				pw.printf("        $this->con->put_block($this->_ooctx);\n");
			} else {
				pw.printf("        $this->con->put_bytes($this->_ooctx);\n");
			}
		}
		if(thisbase != null) {
			pw.printf("        $this->con->put_block($this);\n");
		}
		for(Arg arg : args.getArgs()) {
			if(arg.getUse() != Usage.Out || arg.getType() == DataType.Block) {
		        pw.printf("        $this->con->put_%s($%s);\n", typeName(arg.getType()), arg.getName());
			}
		}
		pw.printf("        $this->con->send();\n");
		pw.printf("        $this->con->recv();\n");
		pw.printf("        $retval = $this->con->get_int32();\n");
		if(ooctx != null) {
			if(!Character.isDigit(ooctx.charAt(0))) {
				pw.printf("        $this->con->get_block($this->_ooctx);\n");
			} else {
				pw.printf("        $this->_ooctx = $this->con->get_bytes();\n");
			}
		}
		if(thisbase != null) {
			pw.printf("        $this->con->get_block($this);\n");
		}
		for(Arg arg : args.getArgs()) {
			if(arg.getUse() != Usage.In) {
				if(arg.getType() != DataType.Block) {
					pw.printf("        $%s = $this->con->get_%s();\n", arg.getName(), typeName(arg.getType()));
				} else {
					//pw.printf("        $%s = new %s();\n", arg.getName(), arg.getImpl());
					pw.printf("        $this->con->get_%s($%s);\n", typeName(arg.getType()), arg.getName());
				}
			}
		}
	}
	@Override
	public void genMethodFooter(Method.ArgList args) {
		pw.printf("        $this->con->close();\n");
		pw.printf("        return $retval;\n");
		pw.printf("    }\n");
	}
	@Override
	public void genClassFooter() {
		pw.printf("}\n");
	}
	@Override
	public void close() {
		pw.printf("\n");
		pw.printf("?>\n");
		pw.close();
	}
}
