/*
 * Copyright 2024 Arne Vajh�j.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.vms.tig.client;

import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;

import dk.vajhoej.vms.tig.Arg;
import dk.vajhoej.vms.tig.Arg.DataType;
import dk.vajhoej.vms.tig.Arg.Usage;
import dk.vajhoej.vms.tig.Method;
import dk.vajhoej.vms.tig.shared.Util;

public class VBNetClientGen implements ClientGen {
	private PrintWriter pw;
	@Override
	public String fileName(String imgnam) {
		return className(imgnam) + ".vb";
	}
	@Override
	public String className(String imgnam) {
		return Util.capitalize(imgnam) + "Client";
	}
	@Override
	public String typeName(Arg.DataType typ) {
		switch(typ) {
			case Byte:
				return "Byte";
			case Word:
				return "Short";
			case LongWord:
				return "Integer";
			case QuadWord:
				return "Long";
			case F_Float:
			case S_Float:
				return "Single";
			case D_Float:
			case G_Float:
			case T_Float:
				return "Double";
			case FixedCharacterString:
			case VariableCharacterString:
			case NullTermCharacterString:
				return "String";
			case Block:
				return "Block";
			default:
				throw new RuntimeException("Unsupported type: " + typ);
		}
	}
	@Override
	public void open(String fnm) throws IOException {
		pw = new PrintWriter(new FileWriter(fnm));
	}
	private boolean usens = false;
	@Override
	public void genClassHeader(String ns, String clznam, String ooctx, String thisbase) {
		pw.printf("' GENERATED CODE - edits may be overwritten\n");
		pw.printf("Imports System\n");
		pw.printf("Imports System.IO\n");
		pw.printf("Imports System.Net.Sockets\n");
		pw.printf("\n");
		if(thisbase != null) {
			pw.printf("Imports Vajhoej.Record\n");
		}
		pw.printf("Imports Vajhoej.VMS.TIG\n");
		pw.printf("\n");
		if(ns != null) {
			pw.printf("NameSpace %s\n", ns);
			pw.printf("\n");
			usens = true;
		}
		if(thisbase != null) {
			pw.printf("<Struct> _\n");
		}
		pw.printf("Public Class %s\n", clznam);
		if(thisbase != null) {
			pw.printf("    Inherits %s\n", Util.capitalize(thisbase));
		}
		if(thisbase != null) {
			pw.printf("    <TransientField> _\n");
		}
		pw.printf("    Private host As String\n");
		if(thisbase != null) {
			pw.printf("    <TransientField> _\n");
		}
		pw.printf("    Private port As Integer\n");
		if(ooctx != null) {
			if(!Character.isDigit(ooctx.charAt(0))) {
				pw.printf("    Private _ooctx As %s = New %s()\n", Util.capitalize(ooctx), Util.capitalize(ooctx));
			} else {
				pw.printf("    Private _ooctx(%d) As Byte\n", Integer.parseInt(ooctx) - 1);
			}
		}
	}
	@Override
	public void genCtor(String clznam, int port, String ooctx, String thisbase) {
		pw.printf("    Public Sub New(host As String)\n");
		pw.printf("        me.host = host\n");
		pw.printf("        me.port = %d\n", port);
		pw.printf("    End Sub\n");
		pw.printf("    Public Sub New()\n");
		pw.printf("        ' just needed for temp instance\n");
		pw.printf("    End Sub\n");
		if(ooctx != null && !Character.isDigit(ooctx.charAt(0))) {
			pw.printf("    Public ReadOnly Property OOContext() As %s\n",  Util.capitalize(ooctx));
			pw.printf("        Get\n");
			pw.printf("            Return _ooctx\n");
			pw.printf("        End Get\n");
			pw.printf("    End Property\n");
		}
	}
	@Override
	public void genMethodHeader(String methnam, Method.ArgList args) {
		pw.printf("    Public Function %s(", Util.capitalize(methnam));
		boolean first = true;
		for(Arg arg : args.getArgs()) {
			if(first) {
				first = false;
			} else {
				pw.printf(", ");
			}
			if(arg.getUse() != Usage.In) {
				// make In and InOut argument ref
				pw.printf("ByRef ");
			}
			pw.printf("%s", arg.getName());
			pw.printf(" As %s", arg.getType() != Arg.DataType.Block ? typeName(arg.getType()) : arg.getImpl());
		}
		pw.printf(") As Integer\n");
		pw.printf("        Dim tc As TcpClient = New TcpClient(host, port)\n");
		pw.printf("        Dim con As Connection = New Connection(tc)\n");
		pw.printf("        con.Prep()\n");
		pw.printf("        con.PutString(\"%s\")\n", methnam);
	}
	@Override
	public void genMethodBody(String ooctx, String thisbase, Method.ArgList args) {
		if(ooctx != null) {
			if(!Character.isDigit(ooctx.charAt(0))) {
				pw.printf("        con.PutBlock(_ooctx)\n");
			} else {
				pw.printf("        con.PutBytes(_ooctx)\n");
			}
		}
		if(thisbase != null) {
			pw.printf("        con.PutBlock(me)\n");
		}
		for(Arg arg : args.getArgs()) {
			if(arg.getUse() != Usage.Out || arg.getType() == DataType.Block) {
		    	// for In and InOut write value to connection
		        pw.printf("        con.Put%s(%s)\n", Util.capitalize(typeName(arg.getType())), arg.getName());
			}
		}
		pw.printf("        con.Send()\n");
		pw.printf("        con.Recv()\n");
		pw.printf("        Dim retval As Integer = con.GetInteger()\n");
		if(ooctx != null) {
			if(!Character.isDigit(ooctx.charAt(0))) {
				pw.printf("        _ooctx = DirectCast(con.GetBlock(Of %s)(GetType(%s)), %s)\n", Util.capitalize(ooctx), Util.capitalize(ooctx), Util.capitalize(ooctx));
			} else {
				pw.printf("        _ooctx = con.GetBytes()\n");
			}
		}
		if(thisbase != null) {
			pw.printf("        con.GetBlock(me)\n");
		}
		for(Arg arg : args.getArgs()) {
			if(arg.getUse() != Usage.In) {
				if(arg.getType() != DataType.Block) {
					// for InOut and Out non-block read value from connection
					pw.printf("        %s = con.Get%s()\n", arg.getName(), Util.capitalize(typeName(arg.getType())));
				} else {
					// for InOut and Out non-block read value from connection (class is needed to properly deserialize)
					pw.printf("        %s = con.Get%s(Of %s)(GetType(%s))\n", arg.getName(), Util.capitalize(typeName(arg.getType())), arg.getImpl(), arg.getImpl());
				}
			}
		}
	}
	@Override
	public void genMethodFooter(Method.ArgList args) {
		pw.printf("        con.Close()\n");
		pw.printf("        Return retval\n");
		pw.printf("    End Function\n");
	}
	@Override
	public void genClassFooter() {
		pw.printf("End Class\n");
		if(usens) {
			pw.printf("End NameSpace\n");
		}
	}
	@Override
	public void close() {
		pw.close();
	}
}
