/*
 * Copyright 2024 Arne Vajh�j.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.vms.tig.server;

import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;

import dk.vajhoej.vms.tig.Arg;
import dk.vajhoej.vms.tig.Arg.DataType;
import dk.vajhoej.vms.tig.Arg.Usage;
import dk.vajhoej.vms.tig.Method;
import dk.vajhoej.vms.tig.shared.Util;

public class JavaServerGen implements ServerGen {
	private PrintWriter pw;
	@Override
	public String fileName(String imgnam) {
		return className(imgnam) + ".java";
	}
	@Override
	public String className(String imgnam) {
		return Util.capitalize(imgnam) + "Server";
	}
	@Override
	public String typeName(Arg.DataType typ) {
		switch(typ) {
			case Byte:
				return "byte";
			case Word:
				return "short";
			case LongWord:
			case F_Float: // we neeed to shuffle bits
				return "int";
			case QuadWord:
			case D_Float: // we neeed to shuffle bits
			case G_Float: // we neeed to shuffle bits
				return "long";
			case S_Float:
				return "float";
			case T_Float:
				return "double";
			case FixedCharacterString:
			case VariableCharacterString:
			case NullTermCharacterString:
				return "String";
			case Block:
				return "Block";
			default:
				throw new RuntimeException("Unsupported type: " + typ);
		}
	}
	private String wrapperName(Arg.DataType typ) {
		switch(typ) {
			case Word:
				return "Word";
			case LongWord:
			case F_Float: // we neeed to shuffle bits
				return "LongWord";
			case QuadWord:
			case D_Float: // we neeed to shuffle bits
			case G_Float: // we neeed to shuffle bits
				return "QuadWord";
			case S_Float:
				return "SFloat";
			case T_Float:
				return "TFloat";
			case FixedCharacterString:
				return "CharacterString";
			case VariableCharacterString:
				return "VariableCharacterString";
			case NullTermCharacterString:
				return "NullTermCharacterString";
			case Block:
				return "ByteBuffer";
			default:
				throw new RuntimeException("Unsupported type: " + typ);
		}
	}
	private String rw(Usage u) {
		switch(u) {
			case In:
				return "readOnly";
			case Out:
				return "writeOnly";
			case InOut:
				return "readWrite";
			default:
				throw new RuntimeException("Unsupported usage: " + u);
		}
	}
	private String pass(Arg arg) {
		return String.format("pass(%swrap).by%s().%s()", arg.getName(), Util.capitalize(arg.getPass().toString()), rw(arg.getUse()));
	}
	private boolean isString(DataType typ) {
		switch(typ) {
			case Word:
			case LongWord:
			case QuadWord:
			case S_Float:
			case T_Float:
			case F_Float:
			case D_Float:
			case G_Float:
			case Block:
				return false;
			case FixedCharacterString:
			case VariableCharacterString:
			case NullTermCharacterString:
				return true;
			default:
				throw new RuntimeException("Unsupported type: " + typ);
		}
	}
	@Override
	public void open(String fnm) throws IOException {
		pw = new PrintWriter(new FileWriter(fnm));
	}
	@Override
	public void genClassHeader(String ns, String clznam, boolean mt, int port) {
		pw.printf("// GENERATED CODE - edits may be overwritten\n");
		if(ns != null) {
			pw.printf("package %s;\n", ns);
			pw.printf("\n");
		}
		pw.printf("import java.io.IOException;\n");
	    pw.printf("import java.net.Socket;\n");
		pw.printf("\n");
		pw.printf("import dk.vajhoej.record.VAXFloatUtil;\n");
		pw.printf("import dk.vajhoej.vms.tig.server.%s;\n", mt ? "ServerMultiThreaded" : "ServerSingleThreaded");
		pw.printf("import dk.vajhoej.vms.tig.shared.Connection;\n");
		pw.printf("\n");
		pw.printf("import static dk.vajhoej.vms.call.VMS.*;\n");
		pw.printf("\n");
		pw.printf("public class %s extends %s {\n", clznam, mt ? "ServerMultiThreaded" : "ServerSingleThreaded");
		pw.printf("    @Override\n");
		pw.printf("    public int getPort() {\n");
		pw.printf("        return %d;\n", port);
		pw.printf("    }\n");
	}
	@Override
	public void genProcessorStart() {
		pw.printf("    @Override\n");
		pw.printf("    public void process(Socket s) throws IOException {\n");
		pw.printf("        Connection con = new Connection(s);\n");
		pw.printf("        con.recv();\n");
		pw.printf("        String target = con.getString();\n");
	}
	@Override
	public void genProcessorFragment(String imgnam, String methnam, String ooctx, String thisbase, Method.ArgList args) {
		pw.printf("        if(target.equals(\"%s\")) {\n", methnam);
		// declare arguments and read from client
		if(ooctx != null || thisbase != null) {
			pw.printf("            byte[] _ooctx = con.getBytes();\n");
		}
		for(Arg arg : args.getArgs()) {
		    if(arg.getType() != DataType.Block) {
                if(arg.getUse() != Usage.Out) {
                	// for In and InOut non-block declare variable and read value from connection
                    pw.printf("            %s %s = con.get%s();\n", typeName(arg.getType()), arg.getName(), Util.capitalize(typeName(arg.getType())));
                } else {
                	// for Out non-block just declare variable
                    pw.printf("            %s %s;\n", typeName(arg.getType()), arg.getName());
                }
            } else {
            	// for Block declare variable and always read value from connection as bytes because we need size
                pw.printf("            byte[] %s = con.getBytes();\n", arg.getName());
            }
		}
		// convert VAX float if needed
		for(Arg arg : args.getArgs()) {
			if(arg.getUse() != Usage.Out) {
				if(arg.getType() == DataType.F_Float) {
					pw.printf("            %s = VAXFloatUtil.s2f(%s);\n", arg.getName(), arg.getName());
				}
				if(arg.getType() == DataType.D_Float) {
					pw.printf("            %s = VAXFloatUtil.t2d(%s);\n", arg.getName(), arg.getName());
				}
				if(arg.getType() == DataType.G_Float) {
					pw.printf("            %s = VAXFloatUtil.t2g(%s);\n", arg.getName(), arg.getName());
				}
			}
		}
		if(ooctx != null || thisbase != null) {
			pw.printf("            ByteBuffer _ooctxwrap = new ByteBuffer(_ooctx);\n");
		}
		// wrap arguments
		for(Arg arg : args.getArgs()) {
			if(arg.getUse() != Usage.Out) {
				// for In and InOut wrap value
				pw.printf("            %s %swrap = new %s(%s);\n", wrapperName(arg.getType()), arg.getName(), wrapperName(arg.getType()), arg.getName());
			} else {
				if(!isString(arg.getType())) {
					// for In non-string wrap default value
					pw.printf("            %s %swrap = new %s();\n", wrapperName(arg.getType()), arg.getName(), wrapperName(arg.getType()));
				} else {
					// for In string wrap 32K long blank string
					pw.printf("            %s %swrap = new %s(32000);\n", wrapperName(arg.getType()), arg.getName(), wrapperName(arg.getType()));
				}
			}
		}
		// call
		pw.printf("            int retval = call(\"%s\", \"%s\"\n", imgnam, methnam);
		if(ooctx != null || thisbase != null) {
			pw.printf("                                     , pass(_ooctxwrap).byReference().readWrite()\n");
		}
		for(Arg arg : args.getArgs()) {
			pw.printf("                                     , %s\n", pass(arg));
		}
		pw.printf("                             );\n");
		// unwrap result
		if(ooctx != null || thisbase != null) {
			pw.printf("            _ooctx = _ooctxwrap.getValue();\n");
		}
		for(Arg arg : args.getArgs()) {
			if(arg.getUse() != Usage.In) {
				if(!isString(arg.getType())) {
					// for InOut and Out non-string unwrap with getValue
					pw.printf("            %s = %swrap.getValue();\n", arg.getName(), arg.getName());
				} else {
					// for InOut and Out non-string unwrap with toString (getValue return a byte array)
					pw.printf("            %s = %swrap.toString();\n", arg.getName(), arg.getName());
				}
			}
		}
		// return value
		pw.printf("            con.prep();\n");
		pw.printf("            con.putInt(retval);\n");
		// convert VAX float if needed
		for(Arg arg : args.getArgs()) {
			if(arg.getUse() != Usage.In) {
				if(arg.getType() == DataType.F_Float) {
					pw.printf("            %s = VAXFloatUtil.f2s(%s);\n", arg.getName(), arg.getName());
				}
				if(arg.getType() == DataType.D_Float) {
					pw.printf("            %s = VAXFloatUtil.d2t(%s);\n", arg.getName(), arg.getName());
				}
				if(arg.getType() == DataType.G_Float) {
					pw.printf("            %s = VAXFloatUtil.g2t(%s);\n", arg.getName(), arg.getName());
				}
			}
		}
		// write arguments back to client
		if(ooctx != null || thisbase != null) {
			pw.printf("            con.putBytes(_ooctx);\n");
		}
		for(Arg arg : args.getArgs()) {
            if(arg.getUse() != Usage.In) {
			    if(arg.getType() != DataType.Block) {
                	// for InOut and Out non-block write value to connection
                    pw.printf("            con.put%s(%s);\n", Util.capitalize(typeName(arg.getType())), arg.getName());
	            } else {
	            	// for InOut and Out block write value to connection as bytes
	                pw.printf("            con.putBytes(%s);\n", arg.getName());
	            }
            }
		}
		pw.printf("        }\n");
	}
	@Override
	public void genProcessorFinish() {
		pw.printf("        con.send();\n");
		pw.printf("    }\n");
	}
	@Override
	public void genClassFooter(String clznam) {
		pw.printf("    public static void main(String[] args) throws IOException {\n");
		pw.printf("         (new %s()).run();\n", clznam);
		pw.printf("    }\n");
		pw.printf("}\n");
	}
	@Override
	public void close() {
		pw.close();
	}
}
