/*
 * Copyright 2024 Arne Vajh�j.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.vms.tig.shared;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.Socket;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.logging.Logger;

import dk.vajhoej.record.RecordException;
import dk.vajhoej.record.StructReader;
import dk.vajhoej.record.StructWriter;

public class Connection {
	private static final int MAX_FRAME_SIZE = 1000000;
	private Socket s;
	private ByteBuffer wrbb;
	private ByteBuffer rdbb;
	private Logger log = Logger.getLogger(Connection.class.getName());
	public Connection(Socket s) {
	    log.fine(Thread.currentThread() + ": ctor s=" + s);
		this.s = s;
	}
	public void prep() {
	    log.fine(Thread.currentThread() + ": prep");
	    byte[] temp = new byte[MAX_FRAME_SIZE];
		wrbb = ByteBuffer.wrap(temp);
		wrbb.order(ByteOrder.LITTLE_ENDIAN);
		log.fine(Thread.currentThread() + ": wrbb=" + wrbb);
		wrbb.putInt(0); // length to be updated later
	}
	public void putByte(byte v) {
	    log.fine(Thread.currentThread() + ": putByte: " + v);
		wrbb.put(v);
	}
	public void putShort(short v) {
	    log.fine(Thread.currentThread() + ": putShort: " + v);
		wrbb.putShort(v);
	}
	public void putInt(int v) {
	    log.fine(Thread.currentThread() + ": putInt: " + v);
		wrbb.putInt(v);
	}
	public void putLong(long v) {
	    log.fine(Thread.currentThread() + ": putLong: " + v);
		wrbb.putLong(v);
	}
	public void putFloat(float v) {
	    log.fine(Thread.currentThread() + ": putFloat: " + v);
		wrbb.putFloat(v);
	}
	public void putDouble(double v) {
	    log.fine(Thread.currentThread() + ": putDouble: " + v);
		wrbb.putDouble(v);
	}
	public void putString(String v) throws UnsupportedEncodingException {
	    log.fine(Thread.currentThread() + ": putString: " + v);
		byte[] temp = v.getBytes("ISO-8859-1");
		wrbb.putShort((short)temp.length);
		wrbb.put(temp);
	}
	public void putBytes(byte[] v) {
		wrbb.putShort((short)v.length);
		wrbb.put(v);
	}
	public void putBlock(Object o) throws RecordException {
	    log.fine(Thread.currentThread() + ": putBlock: " + o);
		StructWriter sw = new StructWriter();
		sw.write(o);
		byte[] temp = sw.getBytes();
		putBytes(temp);
	}
	public void send() throws IOException {
		log.fine(Thread.currentThread() + ": wrbb=" + wrbb);
		int frlen = wrbb.position() - 4;
		wrbb.putInt(0, frlen); // update length
		wrbb.rewind();
		byte[] fr = new byte[4 + frlen];
		wrbb.get(fr);
		log.fine(Thread.currentThread() + ": send frlen=" + frlen);
		OutputStream os = s.getOutputStream();
		os.write(fr);
		os.flush();
	}
	public void recv() throws IOException {
	    log.fine(Thread.currentThread() + ": recv");
	    InputStream is = s.getInputStream();
	    int ix;
	    byte[] temp = new byte[4];
	    ix = 0;
	    while(ix < 4) {
	    	ix += is.read(temp, ix, 4 - ix);
	    }
		rdbb = ByteBuffer.wrap(temp);
		rdbb.order(ByteOrder.LITTLE_ENDIAN);
		log.fine(Thread.currentThread() + ": rdbb=" + rdbb);
		int frlen = rdbb.getInt();
		log.fine(Thread.currentThread() + ": frlen=" + frlen);
		byte[] fr = new byte[frlen];
		ix = 0;
		while(ix < frlen) {
			ix += is.read(fr, ix, frlen - ix);
		}
		rdbb = ByteBuffer.wrap(fr);
		rdbb.order(ByteOrder.LITTLE_ENDIAN);
		log.fine(Thread.currentThread() + ": rdbb=" + rdbb);
	}
	public byte getByte() {
	    log.fine(Thread.currentThread() + ": getByte");
		return rdbb.get();
	}
	public short getShort() {
	    log.fine(Thread.currentThread() + ": getShort");
		return rdbb.getShort();
	}
	public int getInt() {
	    log.fine(Thread.currentThread() + ": getInt");
		return rdbb.getInt();
	}
	public long getLong() {
	    log.fine(Thread.currentThread() + ": getLong");
		return rdbb.getLong();
	}
	public float getFloat() {
	    log.fine(Thread.currentThread() + ": getFloat");
		return rdbb.getFloat();
	}
	public double getDouble() {
	    log.fine(Thread.currentThread() + ": getDouble");
		return rdbb.getDouble();
	}
	public String getString() throws UnsupportedEncodingException {
   	    log.fine(Thread.currentThread() + ": getString");
		int len = rdbb.getShort();
		log.fine(Thread.currentThread() + ": len=" + len);
		byte[] temp = new byte[len];
		rdbb.get(temp);                                                    
		String res = new String(temp, "ISO-8859-1"); 
		log.fine(Thread.currentThread() + ": string=" + res); 
		return res;
	}
	public byte[] getBytes() {
		int len = rdbb.getShort();
		byte[] temp = new byte[len];
		rdbb.get(temp);
		return temp;
	}
	public <T> T getBlock(Class<T> t) throws RecordException {
	    log.fine(Thread.currentThread() + ": getBlock " + t.getName());
	    byte[] temp = getBytes();
		StructReader sr = new StructReader(temp);
		return sr.read(t);
	}
	public <T> void getBlock(T o) throws RecordException {
	    byte[] temp = getBytes();
		StructReader sr = new StructReader(temp);
		sr.read(o);
	}
	public void close() throws IOException {
	    log.fine(Thread.currentThread() + ": close");
		s.close();
	}
}
